#include <util/application.h>
#include <util/message.h>
#include <gui/button.h>
#include <gui/textview.h>
#include <gui/control.h>
#include <atheos/threads.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "syl_window.h"
#include "structs.h"
#include "debugmacros.h"
#include "syl_generic.h"

os::Application *app;

extern "C" {


void app_thread(void * param)
{
	DEBUGPUTS("Before app->Run()\n");
	app->Run();
	DEBUGPUTS("After app->Run()\n");
}

//VIS_CRITICAL_SECTION_EXTERN(app_lock)
extern short app_lock;

void check_create_app()
{
	VIS_EnterCriticalSection(app_lock);
		if(!app)
		{
			
			app = new os::Application("mime type goes here");
			resume_thread(spawn_thread("vis_app", (void *)app_thread, 1, 0, NULL));
			DEBUGPUTS("Created application object\n");
		}
	VIS_LeaveCriticalSection(app_lock);
}

int vis_window_get_value(datum ** inputlist, queue_entry * worker_entry)
{
	datum * output = NULL;
	os::String val;
	os::Control * control = NULL;
	vis_window_shown * window = (vis_window_shown *)(inputlist[0]->c.generic.data);
	DEBUGPRINTF("inputlist[0]: %X\n", inputlist[0]);
	DEBUGPRINTF("window: %X\n", window);
	DEBUGPRINTF("window->title: %X\n", window->title);
	DEBUGPUTS("Entering window lock\n");
	VIS_EnterCriticalSection(window->lock);
		DEBUGPUTS("Before FindView\n");
		DEBUGPRINTF("window->wind: %X\n", window->wind);
		DEBUGPRINTF("Looper: IsLocked: %d, LockCount: %d\n", window->wind->IsLocked(), window->wind->GetLockCount());
		DEBUGPRINTF("Looper: LockingThread: %d\n", window->wind->GetLockingThread());
		control = (os::Control *)(window->wind->FindView((char *)(inputlist[1]->c.generic.data)));
		if(control)
		{
			DEBUGPUTS("Before GetValue\n");
			val = control->GetValue().AsString();
			output = new_datum(BUILTIN_TYPE_STRING, 1, val.CountChars()+1, worker_entry->instance->def->prog);
			strcpy((char *)(output->c.generic.data), val.c_str());
		}
		DEBUGPRINTF("window->wind: %X\n", window->wind);
	VIS_LeaveCriticalSection(window->lock);
	DEBUGPUTS("Leaving window lock\n");
	DEBUGPRINTF("window->wind: %X\n", window->wind);
	DEBUGPRINTF("window->title: %X\n", window->title);
	release_ref(inputlist[0]);
	release_ref(inputlist[1]);
	inputlist[0] = output;
	return 0;
}

int vis_window_set_value(datum ** inputlist, queue_entry * worker_entry)
{
	os::Control * control = NULL;
	vis_window_shown * window = (vis_window_shown *)(inputlist[0]->c.generic.data);
	DEBUGPRINTF("inputlist[0]: %X\n", inputlist[0]);
	DEBUGPRINTF("window: %X\n", window);
	DEBUGPRINTF("window->title: %X\n", window->title);
	DEBUGPUTS("Set Value: Entering window lock\n");
	VIS_EnterCriticalSection(window->lock);
		DEBUGPUTS("Before FindView\n");		
		DEBUGPRINTF("window->wind: %X\n", window->wind);
		DEBUGPRINTF("Looper: IsLocked: %d, LockCount: %d\n", window->wind->IsLocked(), window->wind->GetLockCount());
		DEBUGPRINTF("Looper: LockingThread: %d\n", window->wind->GetLockingThread());
		control = (os::Control *)(window->wind->FindView((char *)(inputlist[1]->c.generic.data)));
		if(control)
		{
			DEBUGPUTS("Before SetValue\n");
			DEBUGPRINTF("Value is: %s\n", inputlist[2]->c.generic.data);
			control->SetValue(os::String((char *)(inputlist[2]->c.generic.data)), false);
			DEBUGPUTS("After SetValue\n");
		}
		DEBUGPRINTF("window->wind: %X\n", window->wind);
	VIS_LeaveCriticalSection(window->lock);
	DEBUGPUTS("Set Value: Leaving window lock\n");
	DEBUGPRINTF("window->wind: %X\n", window->wind);
	DEBUGPRINTF("window->title: %X\n", window->title);
	DEBUGPRINTF("inputlist[0]->refcount: %d\n", inputlist[0]->ref_count);
	release_ref(inputlist[0]);
	release_ref(inputlist[1]);
	release_ref(inputlist[2]);
	return 0;
}

void add_widget_shared(vis_window_shown * window_show, datum * widget_datum, char * name, double xpos, double ypos, queue_entry * worker_entry, os::View * parent_view = NULL)
{
	os::View * widget_view;
	os::TextView * text_view;
	vis_widget * widget;
	datum * params[2];
	widget = (vis_widget *)(widget_datum->c.generic.data);
		DEBUGPUTS("Instantiating OS native widget object\n");
		switch(widget_datum->company->type_id)
		{
			case BUILTIN_TYPE_BUTTON:
				DEBUGPUTS("new os::Button()\n");
				widget_view = new os::Button(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					), 
					name, 
					(char *)(widget->label->c.generic.data), 
					new os::Message(window_show->next_msg_code++)
				);
				params[0] = add_ref(widget->handler_dict);
				params[1] = new_datum(BUILTIN_TYPE_STRING, 1, 6, worker_entry->instance->def->prog);//click
				strcpy((char *)(params[1]->c.generic.data), "click");
				DEBUGPUTS("Looking up click handler\n");
				vis_dict_index(params, worker_entry);
				if(params[1])
					release_ref(params[1]);
				DEBUGPRINTF("Appending %X to handler list\n", params[0]);
				params[1] = params[0];
				params[0] = window_show->handler_list;
				vis_list_append(params, worker_entry);
				window_show->handler_list = params[0];
				break;
			case BUILTIN_TYPE_INPUTBOX:
				text_view = new os::TextView(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					),
					name, 
					(char *)(widget->value->c.generic.data)
				);
				switch(widget->flags)
				{
				case 1:
					text_view->SetMultiLine();
					break;
				case 2:
					text_view->SetNumeric(true);
					break;
				case 3:
					text_view->SetPasswordMode();
					break;
				case 4:
					text_view->SetReadOnly();
					break;
				default:
					break;
				}
				widget_view = text_view;
				break;
			case BUILTIN_TYPE_CUSTOM_WIDGET:
				DEBUGPUTS("Creating custom widget\n");
				widget_view = new GenericView(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					),
					name,
					add_ref(widget_datum),
					&(window_show->instance));
				DEBUGPRINTF("Widget view: %X\n", widget_view);
				break;
			default:
				//Do error crap here
				break;
		}
		if(parent_view) {
			//widget_view->SetFlags(widget_view->GetFlags() | os::WID_TRANSPARENT);
			parent_view->AddChild(widget_view);
		} else {
			DEBUGPUTS("Adding OS native object to window\n");
			window_show->wind->AddChild(widget_view);
		}
	
}

int vis_window_shown_addwidget(datum ** inputlist, queue_entry * worker_entry)
{
	vis_window_shown * window = (vis_window_shown *)(inputlist[0]->c.generic.data);
	window->wind->Lock();
	add_widget_shared(window, inputlist[2], (char *)(inputlist[1]->c.generic.data), inputlist[3]->c.real, inputlist[4]->c.real, worker_entry, NULL);
	window->wind->Unlock();
	release_ref(inputlist[1]);
	release_ref(inputlist[2]);
	release_ref(inputlist[3]);
	release_ref(inputlist[4]);
	return 0;
}

int vis_window_show(datum ** inputlist, queue_entry * worker_entry)
{
	vis_window * window = (vis_window *)(inputlist[0]->c.generic.data);
	vis_window_shown * window_show;
	list_data * list;
	int i;
	datum * params[2];
	double xpos, ypos;
	datum * widget_datum;
	vis_widget * widget;
	os::View * widget_view;
	os::TextView * text_view;
	datum * show_datum = new_datum(BUILTIN_TYPE_WINDOW_SHOWN, 1, sizeof(vis_window_shown), worker_entry->instance->def->prog);
	window_show = (vis_window_shown *)(show_datum->c.generic.data);
	window_show->title = add_ref(window->title);
	window_show->width = window->width;
	window_show->height = window->height;
	window_show->xpos = inputlist[1]->c.real;
	release_ref(inputlist[1]);
	window_show->ypos = inputlist[2]->c.real;
	release_ref(inputlist[2]);
	window_show->is_open = TRUE;
	window_show->wait_entry = NULL;
	window_show->next_msg_code = 0;
	window_show->instance.def = worker_entry->instance->def;
	
	check_create_app();
	VIS_InitializeCriticalSection(window_show->lock);
	window_show->wind = new os_window(add_ref(show_datum));
	
	window_show->handler_list = create_list(worker_entry->instance->def->prog);
	list = (list_data *)(window->id_list->c.generic.data);
	for(i = 0; i < list->num_entries; ++i)
	{
		DEBUGPRINTF("Retrieving widget %d\n", i);
		params[0] = add_ref(window->widget_dict);
		params[1] = add_ref(list->entries[i]);
		vis_dict_index(params, NULL);
		widget_datum = params[0];
		DEBUGPRINTF("Retrieving xpos for widget: %d\n", i);
		params[0] = add_ref(window->widget_xpos);
		params[1] = add_ref(list->entries[i]);
		DEBUGPUTS("Calling vis_dict_index(params, NULL) for xpos\n");
		vis_dict_index(params, NULL);
		DEBUGPUTS("After vis_dict_index\n");
		xpos = params[0]->c.real;
		DEBUGPUTS("Releasing xpos datum\n");
		release_ref(params[0]);
		DEBUGPRINTF("Retrieving ypos for widget: %d\n", i);
		params[0] = add_ref(window->widget_ypos);
		params[1] = add_ref(list->entries[i]);
		DEBUGPUTS("Calling vis_dict_index(params, NULL) for ypos\n");
		vis_dict_index(params, NULL);
		DEBUGPUTS("After vis_dict_index\n");
		ypos = params[0]->c.real;
		DEBUGPUTS("Releasing ypos datum\n");
		release_ref(params[0]);
		add_widget_shared(window_show, widget_datum, (char *)(list->entries[i]->c.generic.data), xpos, ypos, worker_entry, NULL);
		/*
		widget = (vis_widget *)(widget_datum->c.generic.data);
		DEBUGPUTS("Instantiating OS native widget object\n");
		switch(widget_datum->company->type_id)
		{
			case BUILTIN_TYPE_BUTTON:
				DEBUGPUTS("new os::Button()\n");
				widget_view = new os::Button(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					), 
					(char *)(list->entries[i]->c.generic.data), 
					(char *)(widget->label->c.generic.data), 
					new os::Message(window_show->next_msg_code++)
				);
				params[0] = add_ref(widget->handler_dict);
				params[1] = new_datum(BUILTIN_TYPE_STRING, 1, 6, worker_entry->instance->def->prog);//click
				strcpy((char *)(params[1]->c.generic.data), "click");
				DEBUGPUTS("Looking up click handler\n");
				vis_dict_index(params, worker_entry);
				if(params[1])
					release_ref(params[1]);
				DEBUGPRINTF("Appending %X to handler list\n", params[0]);
				params[1] = params[0];
				params[0] = window_show->handler_list;
				vis_list_append(params, worker_entry);
				window_show->handler_list = params[0];
				break;
			case BUILTIN_TYPE_INPUTBOX:
				text_view = new os::TextView(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					),
					(char *)(list->entries[i]->c.generic.data), 
					(char *)(widget->value->c.generic.data)
				);
				switch(widget->flags)
				{
				case 1:
					text_view->SetMultiLine();
					break;
				case 2:
					text_view->SetNumeric(true);
					break;
				case 3:
					text_view->SetPasswordMode();
					break;
				case 4:
					text_view->SetReadOnly();
					break;
				default:
					break;
				}
				widget_view = text_view;
				break;
			case BUILTIN_TYPE_CUSTOM_WIDGET:
				DEBUGPUTS("Creating custom widget\n");
				widget_view = new GenericView(
					os::Rect(
						xpos, ypos, 
						xpos+widget->width, ypos+widget->height
					),
					(char *)(list->entries[i]->c.generic.data),
					add_ref(widget_datum),
					&(window_show->instance));
				DEBUGPRINTF("Widget view: %X\n", widget_view);
				break;
			default:
				//Do error crap here
				break;
		}
		DEBUGPUTS("Adding OS native object to window\n");
		window_show->wind->AddChild(widget_view);*/
	}

	DEBUGPUTS("Showing window\n");
	window_show->wind->MakeFocus();
	window_show->wind->Show();
	//resume_thread(spawn_thread("vis_mem_check", (void *)vis_mem_check, 1, 0, (void *)window_show));
	//sleep(5);
	release_ref(inputlist[0]);
	inputlist[0] = show_datum;
	return 0;
}


}

os_window::os_window(datum * vis_wind) : os::Window(
											os::Rect(
												((vis_window_shown *)(vis_wind->c.generic.data))->xpos, 
												((vis_window_shown *)(vis_wind->c.generic.data))->ypos, 
												((vis_window_shown *)(vis_wind->c.generic.data))->xpos+((vis_window_shown *)(vis_wind->c.generic.data))->width,
												((vis_window_shown *)(vis_wind->c.generic.data))->ypos+((vis_window_shown *)(vis_wind->c.generic.data))->height), 
											"vis_wnd", 
											(char *)(((vis_window_shown *)(vis_wind->c.generic.data))->title->c.generic.data))
{
	//DEBUGPRINTF("vis_wind lock: %d\n", ((vis_window_shown *)(vis_wind->c.generic.data))->lock);
	this->vis_wind = vis_wind;
//	this->current_focus = NULL;
}
os_window::~os_window()
{
	vis_window_closed(this->vis_wind);
}

void os_window::HandleMessage(os::Message * msg)
{
	queue_entry entry;
	datum * params[3];
	int code = msg->GetCode();
	DEBUGPRINTF("Message Code: %d\n", code);
	DEBUGPRINTF("window->wind: %X\n", ((vis_window_shown *)(this->vis_wind->c.generic.data))->wind);
	DEBUGPRINTF("vis_wind->ref_count: %X\n", this->vis_wind->ref_count);
	params[0] = add_ref(((vis_window_shown *)(this->vis_wind->c.generic.data))->handler_list);
	params[1] = new_datum(BUILTIN_TYPE_WHOLE, 2, 0, ((vis_window_shown *)(this->vis_wind->c.generic.data))->instance.def->prog);
	params[1]->c.integers.num_a = code;
	vis_list_index(params, NULL);
	if(params[0] && params[0]->company->type_id == BUILTIN_TYPE_WORKER)
	{
		entry.worker_num = 0;
		entry.instance = &(((vis_window_shown *)(this->vis_wind->c.generic.data))->instance);
		params[1] = create_list(entry.instance->def->prog);
		params[2] = add_ref(this->vis_wind);
		vis_list_append(params + 1, &entry);
		vis_worker_do(params, &entry);
		//execute_def_wait(deflist + params[0]->c.integers.num_a, params+1);
		//execute_def(((vis_window_shown *)(this->vis_wind->c.generic.data))->instance.def->prog->deflist + params[0]->c.integers.num_a, entry, params+1, NULL);
	}
	DEBUGPRINTF("window->wind: %X\n", ((vis_window_shown *)(this->vis_wind->c.generic.data))->wind);
	DEBUGPUTS("End: HandleMessage\n");
}
/*
void os_window::DispatchMessage(os::Message * pcMsg, os::Handler * pcHandler)
{
	os::View * view;
	os::View * handler_view;
	//int code = pcMsg->GetCode();
	pcMsg->FindPointer("_widget", (void **)&view);
	if(view) {
		handler_view = _FindHandler(view->m_nToken);
		if(handler_view != view)
			printf("DispatchMessage, message view: %X, handler view: %X\n", view, handler_view);
		
	}
	os::Window::DispatchMessage(pcMsg, pcHandler);
}

os::View *os_window::SetFocusNoHandler(os::View * new_focus )
{
	ifthis->current_focus)
		this->current_focus->Activated(false);
	if(new_focus)
		new_focus->Activated(true);
	this->current_focus = new_focus;
	return new_focus;
}*/



















