#include "structs.h"
#include "interp.h"
#include "datum.h"
#include "parser.h"
#include "saveload.h"
#include <stdlib.h>

int vis_worker_from_string(datum ** inputlist, queue_entry * worker_entry)
{
	int i;
	worker_datum * work;
	datum * output = new_datum(BUILTIN_TYPE_WORKER, 1, sizeof(worker_datum), worker_entry->instance->def->program);
	work = output->c.generic.data;
	find_worker(inputlist[0]->c.generic.data, NULL, NULL, worker_entry->instance->def->program, &work->def);
	for(i = 0; i < 32; ++i)
		work->params[i] = NULL;
	release_ref(inputlist[0]);
	inputlist[0] = output;
	return 0;
}

BOOL list_to_array(datum * list_dat, datum ** out_params, int num)
{
	int i;
	datum * params[2];
	list_data * list = list_dat->c.generic.data;
	if(num > list->num_entries) {
		return FALSE;
	}
	for(i = 0; i < num; ++i)
		out_params[i] = add_ref(list->entries[i]);
	release_ref(list_dat);
	return TRUE;
}

datum * array_to_list(datum ** in_params, int num, program * prog)
{
	int i;
	datum * params[2];
	params[0] = create_list(prog);
	for(i = 0; i < num; ++i)
	{
		params[1] = in_params[i];
		vis_list_append(params, NULL);
	}
	return params[0];
}

int worker_populate_inputs(datum * inputWorker, datum * ioList, datum ** inputlist)
{
	int i,j=0;
	worker_datum * work = inputWorker->c.generic.data;
	list_data * list = ioList->c.generic.data;
	for(i = 0; i < work->def->num_inputs; ++i)
		if(work->params[i])
		{
			DEBUGPRINTF("Setting input %d to %X from worker object\n", i, work->params[i]);
			inputlist[i] = add_ref(work->params[i]);
		}
		else
		{
			if(j < list->num_entries)
			{
				DEBUGPRINTF("Setting input %d to %X from input %d\n", i, list->entries[j], j);
				inputlist[i] = add_ref(list->entries[j]);
				++j;
			}
			else
			{
				ERRORPRINTF("Error: List passed to Do@Worker doesn't have enough entries for worker %s\n", work->def->name);
				return -1;
			}
		}
	return 0;
}

int vis_worker_do(datum ** inputlist, queue_entry * worker_entry)
{
	//int i,j;
	worker_datum * work;
	//worker_def * def;
	//list_data * list;
	datum * inputWorker = inputlist[0];
	datum * ioList = inputlist[1];
	int returnval;
	//release_ref(inputlist[0]);
	work = inputlist[0]->c.generic.data;
	//list_to_array(inputlist[1], inputlist, work->def->num_inputs);
	/*list = inputlist[1]->c.generic.data;
	j = 0;
	DEBUGPRINTF("vis_worker_do: def_name: %s, num_inputs = %d, list length = %d\n", work->def->name, work->def->num_inputs, list->num_entries);
	for(i = 0; i < work->def->num_inputs; ++i)
		if(work->params[i])
		{
			DEBUGPRINTF("Setting input %d to %X from worker object\n", i, work->params[i]);
			inputlist[i] = add_ref(work->params[i]);
		}
		else
		{
			DEBUGPRINTF("Setting input %d to %X from input %d, company name: %s\n", i, list->entries[j], j, list->entries[j]->company->name);
			inputlist[i] = add_ref(list->entries[j]);
			++j;
		}*/
	returnval = worker_populate_inputs(inputWorker, ioList, inputlist);
	release_ref(ioList);
	if(returnval)
		return returnval;
	returnval = execute_def(work->def, *worker_entry, inputlist, pack_list_sub_callback);
	if(returnval == 0)
	{
		inputlist[0] = array_to_list(inputlist, work->def->num_outputs, worker_entry->instance->def->program);
		/*vis_list_new(params, NULL);
		DEBUGPRINTF("params[0] after vis_list_new: %X\n", params[0]);
		for(i = 0; i < work->def->num_outputs; ++i)
		{
			params[1] = inputlist[i];
			vis_list_append(params, NULL);
			DEBUGPRINTF("params[0] after vis_list_append: %X\n", params[0]);
		}
		inputlist[0] = params[0];*/
	}
	release_ref(inputWorker);
	return returnval;
}

int vis_worker_setinput(datum ** inputlist, queue_entry * worker_entry)
{
	worker_datum * work;
	inputlist[0] = copy_datum(inputlist[0], 0);
	work = inputlist[0]->c.generic.data;
	if(inputlist[1]->c.integers.num_a < 32)
		work->params[inputlist[1]->c.integers.num_a] = inputlist[2];
	else
	{
		puts("Error: Visuality currently only supports 32 inputs\n");
		execute_active = FALSE;
		return -1;
	}
	release_ref(inputlist[1]);
	inputlist[1] = inputlist[2] = NULL;
	return 0;
}

int vis_worker_add_worker_call(datum ** inputlist, queue_entry * worker_entry)
{
	worker_datum * add_work;
	worker_datum * work = inputlist[0]->c.generic.data;
	int index;
	add_work = inputlist[1]->c.generic.data;
	/*index = add_work->def - work->def->program->deflist;
	DEBUGPRINTF("work->def: %X, deflist: %X\n", work->def, worker_entry->instance->def->program->deflist);
	DEBUGPRINTF("work->def->name: %X\n", work->def->name);*/
	index = add_worker_to_def(work->def, add_work->def, 1.0, 1.0);
	release_ref(inputlist[1]);
	inputlist[1] = new_datum(BUILTIN_TYPE_WHOLE, 2, 0, worker_entry->instance->def->program);
	inputlist[1]->c.integers.num_a = index;
	return 0;
}

int vis_worker_add_wire(datum ** inputlist, queue_entry * worker_entry)
{
	worker_datum * work = inputlist[0]->c.generic.data;
	add_wire(work->def, inputlist[1]->c.integers.num_a, inputlist[2]->c.integers.num_a, inputlist[3]->c.integers.num_a, inputlist[4]->c.integers.num_a);
	release_ref(inputlist[1]);
	release_ref(inputlist[2]);
	release_ref(inputlist[3]);
	release_ref(inputlist[4]);
	return 0;
}

int vis_worker_add_constant(datum ** inputlist, queue_entry * worker_entry)
{
	datum * constant = inputlist[1];
	worker_datum * work = inputlist[0]->c.generic.data;
	int index = add_constant(work->def, "code generated", 1.0, 1.0);
	if(worker_entry->instance->def->program != work->def->program && constant->company->type_id < USER_DEFINED_TYPES)
	{
		constant = copy_datum(constant, 0);
		constant->company = work->def->program->companylist + constant->company->type_id;
	}
	work->def->implement_func->workerlist[index].value_index = constant;
	inputlist[1] = new_datum(BUILTIN_TYPE_WHOLE, 2, 0, worker_entry->instance->def->program);
	inputlist[1]->c.integers.num_a = index;
	return 0;
}

int vis_worker_add_input(datum ** inputlist, queue_entry * worker_entry)
{
	worker_datum * work = inputlist[0]->c.generic.data;
	int index = add_input_num(work->def, inputlist[1]->c.generic.data, inputlist[2]->c.integers.num_a, 1.0, 1.0);
	release_ref(inputlist[1]);
	release_ref(inputlist[2]);
	inputlist[1] = new_datum(BUILTIN_TYPE_WHOLE, 2, 0, worker_entry->instance->def->program);
	inputlist[1]->c.integers.num_a = index;
	return 0;
}

int vis_worker_add_output(datum ** inputlist, queue_entry * worker_entry)
{
	worker_datum * work = inputlist[0]->c.generic.data;
	int index = add_output_num(work->def, inputlist[1]->c.generic.data, inputlist[2]->c.integers.num_a, 1.0, 1.0);
	release_ref(inputlist[1]);
	release_ref(inputlist[2]);
	inputlist[1] = new_datum(BUILTIN_TYPE_WHOLE, 2, 0, worker_entry->instance->def->program);
	inputlist[1]->c.integers.num_a = index;
	return 0;
}

int vis_worker_new(datum ** inputlist, queue_entry * worker_entry)
{
	int i;
	datum * output = new_datum(BUILTIN_TYPE_WORKER, 1, sizeof(worker_datum), worker_entry->instance->def->program);
	worker_datum * work = output->c.generic.data;
	work->def = create_worker(worker_entry->instance->def->program, inputlist[0]->c.generic.data, 0, 0, USER_FLAG | WORKER_TYPE);
	for(i = 0; i < 32; ++i)
		work->params[i] = NULL;
	release_ref(inputlist[0]);
	inputlist[0] = output;
	return 0;
}

int vis_program_new(datum ** inputlist, queue_entry * worker_entry)
{
	program * prog = new_program(START_DEF_STORAGE, START_COMP_STORAGE);
	inputlist[0] = new_datum(BUILTIN_TYPE_PROGRAM, 2, 0, worker_entry->instance->def->program);
	inputlist[0]->c.generic.data = prog;
	inputlist[0]->c.generic.len = sizeof(program);
	inputlist[0]->union_type = 1;
	return 0;
}

int vis_program_new_worker(datum ** inputlist, queue_entry * worker_entry)
{
	int i;
	datum * output = new_datum(BUILTIN_TYPE_WORKER, 1, sizeof(worker_datum), worker_entry->instance->def->program);
	worker_datum * work = output->c.generic.data;
	work->def = create_worker(inputlist[0]->c.generic.data, inputlist[1]->c.generic.data, 0, 0, USER_FLAG | WORKER_TYPE);
	for(i = 0; i < 32; ++i)
		work->params[i] = NULL;
	release_ref(inputlist[1]);
	inputlist[1] = output;
	return 0;
}

int vis_program_add_worker(datum ** inputlist, queue_entry * worker_entry)
{
	int i;
	worker_datum * work = inputlist[1]->c.generic.data;
	program * prog = inputlist[0]->c.generic.data;
	datum * output = new_datum(BUILTIN_TYPE_WORKER, 1, sizeof(worker_datum), worker_entry->instance->def->program);
	worker_datum * out_work = output->c.generic.data;
	
	out_work->def = create_worker(prog, work->def->name, work->def->num_inputs, work->def->num_outputs, WORKER_TYPE);
	out_work->def->implement_func = work->def->implement_func;
	out_work->def->type = work->def->type;
	if(work->def->type & USER_FLAG)
		out_work->def->program = work->def->program;
	for(i = 0; i < work->def->num_inputs; ++i)
		out_work->def->input_types[i] = out_work->def->input_types[i];
	release_ref(inputlist[1]);
	inputlist[1] = output;
	return 0;
}

int vis_program_add_builtins(datum ** inputlist, queue_entry * worker_entry)
{
	initpredefworkers((program *)inputlist[0]->c.generic.data);
	return 0;
}

int vis_program_run(datum ** inputlist, queue_entry * worker_entry)
{
	datum * params[32];
	params[0] = inputlist[1];
	VIS_EnterCriticalSection(program_count_lock);
		++program_count;
	VIS_LeaveCriticalSection(program_count_lock);
	init_custom_worker(-1, NULL, ((program *)(inputlist[0]->c.generic.data))->defs->deflist, main_callback, add_ref(inputlist[0]), params);
	return 0;
}

int vis_program_find_worker(datum ** inputlist, queue_entry * worker_entry)
{
	int i;
	worker_datum * work;
	program * prog = inputlist[0]->c.generic.data;
	datum * output = new_datum(BUILTIN_TYPE_WORKER, 1, sizeof(worker_datum), worker_entry->instance->def->program);
	work = output->c.generic.data;
	find_worker(inputlist[1]->c.generic.data, NULL, NULL, prog, &(work->def));
	for(i = 0; i < 32; ++i)
		work->params[i] = NULL;
	release_ref(inputlist[1]);
	release_ref(inputlist[0]);
	inputlist[0] = output;
	return 0;
}
/*
int vis_worker_begin_transaction(datum ** inputlist, queue_entry * worker_entry)
{
	worker_def * def;
	int returnval;
	datum * params[2];
	int i;
	global_store * store;
	list_data * list = inputist[0]->c.generic.data;
	transaction * trans = malloc(sizeof(transaction) + list->num_entries - 1);
	trans->num_stores = list->num_entries;
	VIS_EnterCriticalSection(global_store_lock);
		for(i = 0; i < list->num_entries; ++i)
		{
			params[0] = global_store_dict;
			params[1] = add_ref(list->entries[i]);
			vis_dict_index(params, NULL);
			store = params[0]->c.generic.data;
			while(store->inuse)
			{
				VIS_LeaveCriticalSection(global_store_lock);
					#ifdef WIN32
						Sleep(0);
					#else
						sleep(0);
					#endif
				VIS_EnterCriticalSection(global_store_lock);
			}
			store->inuse;
			trans->stores[i] = params[0];
		}
	VIS_LeaveCriticalSection(global_store_lock;
	release_ref(inputlist[0]);
	def = deflist+inputlist[1]->c.integers.num_a;
	release_ref(inputlist[1]);
	list_to_array(inputlist[2], inputlist, def->num_inputs);
	returnval = execute_def_data(def, *worker_entry, inputlist, trasnaction_sub_callback, transaction);
	if(returnval == 0)
		inputlist[0] = array_to_list(inputlist, def->num_outputs);
	return returnval;
}

*/

